import numpy as np
import torch


def get_non_pad_mask(src_pos):
    return (src_pos > 0).type(torch.float).unsqueeze(-1)


def get_sinusoid_encoding_table(num_positions, hidden_size, padding_idx=None):
    """Sinusoid position encoding table"""

    def cal_angle(position, hid_idx):
        return position / np.power(10000, 2 * (hid_idx // 2) / hidden_size)

    def get_posi_angle_vec(position):
        return [cal_angle(position, hid_j) for hid_j in range(hidden_size)]

    sinusoid_table = np.array([get_posi_angle_vec(pos_i) for pos_i in range(num_positions)])

    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    if padding_idx is not None:
        # zero vector for padding dimension
        sinusoid_table[padding_idx] = 0.

    return torch.FloatTensor(sinusoid_table)


def get_attn_key_pad_mask(seq_pos, seq):
    """For masking out the padding part of key sequence."""
    # Expand to fit the shape of key query attention matrix.
    len_seq = seq.size(1)
    padding_mask = seq_pos.eq(0)
    padding_mask = padding_mask.unsqueeze(1).expand(-1, len_seq, -1)  # b x lq x lk
    return padding_mask


def get_subsequent_mask(seq):
    """For masking out the subsequent info."""
    sz_b, len_s = seq.size()
    subsequent_mask = torch.triu(
        torch.ones((len_s, len_s), device=seq.device, dtype=torch.uint8), diagonal=1)
    subsequent_mask = subsequent_mask.unsqueeze(0).expand(sz_b, -1, -1)  # b x ls x ls

    return subsequent_mask